# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
import os
import threading
import time
from queue import Empty
from typing import Any
from typing import Coroutine
from typing import Iterable

from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IMessageHandler
from aorta.types import IPool


class Worker(threading.Thread):
    interval: float = 1
    logger: logging.Logger = logging.getLogger('aorta.transport')
    must_exit: bool = False
    tasks: set[asyncio.Task[None]]

    def __init__(self, pool: IPool, limit: int):
        self.limit = limit
        self.pool = pool
        self.worker_id = bytes.hex(os.urandom(24))
        super().__init__(
            target=lambda: asyncio.run(self.main_event_loop()),
            daemon=True
        )
        self.start()

    def create_task(self, c: Coroutine[Any, Any, Any]):
        if self.loop.is_closed():
            self.logger.critical(
                "Received a task on closed event loop."
            )
            return
        task = self.loop.create_task(c)
        self.tasks.add(task)
        return task

    def configure(self) -> None:
        self.loop = asyncio.get_event_loop()
        self.tasks = set()

    def join(self, timeout: float | None = None):
        self.stop()

    def main_event(self) -> None:
        if len(self.tasks) >= self.limit:
            return
        while len(self.tasks) < self.limit:
            try:
                frame, envelope, handlers = self.pool.get()
                self.create_task(self.execute(frame, envelope, handlers))
            except Empty:
                break

    def on_task_completed(
        self,
        task: asyncio.Task[Any],
        envelope: Envelope[Any],
    ) -> None:
        pass

    def stop(self):
        self.must_exit = True
        self.logger.info("Tearing down worker (id: %s)", self.ident)
        return self

    async def main_event_loop(self):
        self.logger.debug("Spawning worker (thread: %s)", self.ident)
        self.publisher = self.pool.get_publisher()
        self.runner = self.pool.get_runner(self.publisher)
        self.configure()
        self.logger.debug("Start polling for new messages (thread: %s)", self.ident)
        while True:
            # Check pending tasks before doing anything.
            try:
                self.main_event()
            except Exception as e:
                self.logger.exception(
                    "Caught fatal %s in main event.",
                    type(e).__name__
                )
            if self.tasks:
                done, *_ = await asyncio.wait(
                    fs=self.tasks,
                    timeout=0.1,
                    return_when='FIRST_COMPLETED'
                )
                for t in done:
                    self.tasks.remove(t)
            if self.must_exit:
                if self.tasks:
                    await asyncio.wait(self.tasks)
                break
            time.sleep(self.interval)

    async def execute(
        self,
        frame: Acknowledgable,
        envelope: Envelope[Any],
        handlers: Iterable[type[IMessageHandler]]
    ) -> None:
        return await self.runner.run(
            publisher=self.publisher,
            envelope=envelope,
            handlers=handlers
        )


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Union

from .commandhandler import CommandHandler
from .eventlistener import EventListener
from .types import Event
from .types import Command


class Ping(Command):
    pass


class Pong(Command):
    pass


class PingPonged(Event):
    pass


CommandType = Union[Ping, Pong]


class PingHandler(CommandHandler[CommandType]):

    async def handle(self, command: CommandType) -> Any:
        if isinstance(command, Ping):
            self.logger.warning(
                "Ping (id: %s, correlation-id: %s)",
                self.metadata.uid, self.metadata.correlation_id
            )
            await self.send(Pong())
        if isinstance(command, Pong):
            self.logger.warning(
                "Pong (id: %s, correlation-id: %s)",
                self.metadata.uid, self.metadata.correlation_id
            )
            self.publish(PingPonged())


class OnPingPonged(EventListener[PingPonged]):

    async def handle(self, event: PingPonged) -> None:
        self.logger.warning(
            "Ping-ponged (id: %s, correlation-id: %s)",
            self.metadata.uid, self.metadata.correlation_id
        )


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import inspect
import json
from types import ModuleType
from typing import Any

from .types import Command
from .types import Event
from .messagehandler import MessageHandler
from .sewer import Sewer
from .types import Envelope
from .types import IMessageHandler
from .types import MessageHeader


class Provider:
    __module__: str = 'aorta'
    __handlers__: dict[tuple[str, str], set[type[IMessageHandler]]] = collections.defaultdict(set)
    __sewers__: set[type[Sewer[Any]]] = set()

    @classmethod
    def register(cls, handler_class: type[MessageHandler[Any]] | ModuleType) -> None:
        """Register a message handler implementation."""
        if isinstance(handler_class, ModuleType):
            members: list[str] | None = getattr(handler_class, '__all__', None)
            for attname, member in inspect.getmembers(handler_class):
                member: MessageHandler[Any] | Any
                if not inspect.isclass(member):
                    continue
                if members and attname not in members:
                    continue
                if not issubclass(member, MessageHandler):
                    continue
                cls.register(member) # type: ignore
            return
        if issubclass(handler_class, Sewer):
            cls.__sewers__.add(handler_class)
            return

        for Message in handler_class.handles:
            k = (getattr(Message, '__version__', 'v1'), Message.__name__)
            cls.__handlers__[k].add(handler_class)

    @classmethod
    def get(cls, envelope: Envelope[Any] | MessageHeader, sewers: bool = True) -> set[type[IMessageHandler]]:
        """Get the set of message handlers that are able to handle the
        message contained in the envelope.
        """
        if not isinstance(envelope, Envelope):
            return set()
        handlers = set(cls.__handlers__[envelope.qualname])
        if sewers:
            handlers |= set(cls.__sewers__)
        return handlers
    
    @classmethod
    def num_sewers(cls) -> int:
        return len(cls.__sewers__)

    @staticmethod
    def loads(buf: str | bytes):
        return Provider.parse(json.loads(buf))


    @staticmethod
    def parse(data: Any) -> Envelope[Any] | MessageHeader | None:
        """Parses a datastructure into a registered message type
        declaration. Return the evelope or ``None``.
        """
        return (
            Event.parse(data) or
            Command.parse(data)
        )


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import json
from typing import Any
from typing import Callable
from typing import ParamSpec

from ._localpublisher import LocalPublisher
from .baserunner import BaseRunner
from .commandhandler import CommandHandler
from .eventlistener import EventListener
from .messagehandler import MessageHandler
from .messagepublisher import MessagePublisher
from .nulltransport import NullTransport
from .runners import DefaultRunner
from .ping import Pong
from .ping import Ping
from .ping import PingHandler
from .ping import OnPingPonged
from .provider import Provider
from .transaction import Transaction
from .sewer import Sewer
from .types import Command
from .types import Event
from .types import HandlerException
from . import types


__all__: list[str] = [
    'get',
    'parse',
    'register',
    'types',
    'BaseRunner',
    'Command',
    'CommandHandler',
    'DefaultRunner',
    'Event',
    'EventListener',
    'HandlerException',
    'MessageHandler',
    'MessagePublisher',
    'NullTransport',
    'Ping',
    'Pong',
    'Provider',
    'Sewer',
    'Transaction',
]

P = ParamSpec('P')


get = Provider.get
register = Provider.register


def loads(buf: str | bytes):
    return parse(json.loads(buf))


def parse(data: Any) -> types.Envelope[Any] | types.MessageHeader | None:
    """Parses a datastructure into a registered message type
    declaration. Return the evelope or ``None``.
    """
    return (
        Event.parse(data) or
        Command.parse(data)
    )


def run(
    cls: Callable[P, Command | Event] | Command | Event,
    publisher: MessagePublisher | None = None,
    *args: P.args,
    **kwargs: P.kwargs
) -> None:
    """Run a message in the current process and thread."""
    created = False
    message = cls
    if callable(cls):
        message = cls(*args, **kwargs)

    assert isinstance(message, (Command, Event))
    envelope = message.envelope()
    runner = DefaultRunner()
    c = runner.run(
        publisher=publisher or LocalPublisher(get, runner),
        envelope=envelope,
        handlers=get(envelope)
    )
    loop = None
    try:
        loop = asyncio.get_running_loop()
        asyncio.run_coroutine_threadsafe(c, loop)
    except RuntimeError:
        created = True
        loop = asyncio.new_event_loop()
        return loop.run_until_complete(c)
    finally:
        if created and loop is not None and not loop.is_closed():
            loop.close()


register(PingHandler)
register(OnPingPonged)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
from typing import Any
from typing import Coroutine

from .messagehandler import MessageHandler
from .transaction import Transaction
from .provider import Provider
from .types import Drop
from .types import Envelope
from .types import IMessageHandler
from .types import IPublisher
from .types import ITransaction


class BaseRunner:
    __module__: str = 'aorta'
    logger: logging.Logger = logging.getLogger('aorta')
    provider: type[Provider] | Provider
    publisher: IPublisher

    async def run(self, envelope: Envelope[Any]) -> tuple[bool, list[Envelope[Any]]]:
        if envelope.is_bound() and envelope.is_known():
            handlers = self.get_handlers(envelope, False)
            assert len(handlers) == 1, len(handlers)
            handler_class = handlers.pop()
            self.logger.debug(
                "Running handler %s for bound envelope (id: %s, kind; %s, version: %s)",
                handler_class.__name__, envelope.metadata.uid, envelope.kind,
                envelope.api_version
            )
            _, success, _ = await self.run_envelope(handler_class, envelope) # type: ignore

            # TODO: The envelope should be republished here for consistent behavior,
            # but in the current implementation with cbra this does not work well with
            # Google Eventarc and automatic backoff.
            return success, []

        futures: list[Coroutine[Any, Any, Any]] = []
        handler_classes = self.get_handlers(envelope)
        for handler_class in handler_classes:
            if not self.must_run(envelope, handler_class):
                continue
            tx = Transaction(self.publisher, metadata=envelope.metadata)
            handler = handler_class(publisher=tx, metadata=envelope.metadata)
            if not await handler.wants(envelope.message):
                continue
            self.logger.debug(
                "Running handler %s for unbound envelope (id: %s, kind; %s, version: %s)",
                handler_class.__name__, envelope.metadata.uid, envelope.kind,
                envelope.api_version
            )
            futures.append(self.run_handler(tx, handler, envelope)) # type: ignore
    
        finished: set[asyncio.Task[Any]] = set()
        running: set[asyncio.Task[Any]] = set()
        if futures:
            finished, running = await asyncio.wait(
                [asyncio.ensure_future(c) for c in futures],
                return_when=asyncio.ALL_COMPLETED
            )
        assert not running
        retry: list[Envelope[Any]] = []
        for task in finished:
            envelope, success, _ = task.result()
            if envelope and not success:
                retry.append(envelope)

        return not bool(retry), retry

    def get_handlers(self, envelope: Envelope[Any], sewers: bool = True) -> set[type[MessageHandler[Any]]]:
        return self.provider.get(envelope, sewers)
    
    def must_run(self, envelope: Envelope[Any], handler_class: Any):
        return envelope.metadata.handler is None or envelope.wants(handler_class)

    async def run_envelope(
        self,
        handler_class: type[MessageHandler[Any]],
        envelope: Envelope[Any]
    ) -> tuple[Envelope[Any] | None, bool, Any]:
        tx = Transaction(self.publisher)
        handler = handler_class(
            publisher=tx,
            metadata=envelope.metadata
        )
        return await self.run_handler(tx, handler, envelope) # type: ignore

    async def run_handler(
        self,
        transaction: ITransaction,
        handler: IMessageHandler,
        envelope: Envelope[Any]
    ) -> tuple[Envelope[Any] | None, bool, Any]:
        try:
            success, result = await self.handle(transaction, handler, envelope) # type: ignore
        except Drop as e:
            success = True
            result = NotImplemented
            self.logger.warning(
                "Dropping %s/%s: %s (id: %s, correlation-id: %s)",
                envelope.api_version,
                envelope.kind,
                e.reason or 'reason unknown',
                envelope.metadata.uid,
                envelope.metadata.correlation_id
            )
        except Exception as e:
            result = NotImplemented
            success = False
            self.logger.exception(
                "Caught fatal %s during %s.handle()",
                type(e).__name__, type(handler).__name__
            )
        return envelope.clone(type(handler)) if not success else None, success, result 
    
    async def handle(
        self,
        transaction: ITransaction,
        handler: IMessageHandler,
        envelope: Envelope[Any]
    ) -> tuple[bool, Any]:
            async with transaction:
                return await handler.run(envelope)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar
from types import UnionType

from .messagehandler import MessageHandler
from .types import Command
from .types import MessageHandlerType


T = TypeVar('T', bound=Command | UnionType)


class CommandHandlerMetaclass(MessageHandlerType):
    handles: type[Command] = Command
    parameter_name: str = 'command'


class CommandHandler(MessageHandler[Command], Generic[T], metaclass=CommandHandlerMetaclass):
    __module__: str = 'aorta'
    __abstract__: bool = True

    async def handle(self, command: T) -> None:
        raise NotImplementedError

    async def wants(self, command: T) -> bool: # type: ignore
        return True


# Copyright (C) 2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import cast
from typing import Any

import fastapi
from aorta.types import Envelope
from .params import MessageRunner

from .params import AortaMessage


class AortaRouter(fastapi.APIRouter):
    logger: logging.Logger = logging.getLogger('aorta.messages')

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self.add_api_route(
            methods=['POST'],
            path='/command',
            endpoint=self.handle_command,
            include_in_schema=False
        )
        self.add_api_route(
            methods=['POST'],
            path='/event',
            endpoint=self.handle_event,
            include_in_schema=False
        )

    async def handle_command(self, runner: MessageRunner, envelope: AortaMessage) -> fastapi.Response:
        status_code: int = 200
        if envelope is not None and envelope.is_command():
            self.logger.info(
                "Received command (kind: %s/%s, id: %s, correlation-id: %s)",
                envelope.api_version,
                envelope.kind,
                envelope.metadata.uid,
                envelope.metadata.correlation_id,
            )
            assert isinstance(envelope, Envelope)
            success, *_ = await runner.run(cast(Envelope[Any], envelope))
            if not success:
                status_code = 500
        return fastapi.Response(status_code=status_code)

    async def handle_event(self, runner: MessageRunner, envelope: AortaMessage) -> fastapi.Response:
        if envelope is not None and envelope.is_event():
            assert isinstance(envelope, Envelope)
            self.logger.info(
                "Received event (kind: %s/%s, id: %s, correlation-id: %s)",
                envelope.api_version,
                envelope.kind,
                envelope.metadata.uid,
                envelope.metadata.correlation_id,
            )
            _, retry = await runner.run(cast(Envelope[Any], envelope))
            for m in retry:
                self.logger.critical(
                    "Handler failure (handler: %s, message: %s)",
                    m.metadata.handler, m.metadata.uid
                )
        return fastapi.Response(status_code=200)


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import get_args
from typing import get_origin
from typing import Annotated
from typing import Any
from typing import Callable
from typing import Generator

import aorta
import fastapi
import fastapi.params
from aorta.types import Drop
from fastapi.exceptions import RequestValidationError
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

from .messagepublisher import MessagePublisher


class EndpointRunner(aorta.LocalRunner):
    __module__: str = 'aorta.ext.fastapi'
    dependency_cache: dict[tuple[Callable[..., Any], tuple[str]], Any]
    request: fastapi.Request

    @classmethod
    def inject(cls) -> Any:
        return fastapi.Depends(cls)

    def __init__(
        self,
        request: fastapi.Request,
        publisher: MessagePublisher
    ):
        super().__init__(publisher=publisher)
        self.dependency_cache = {}
        self.request = request

    async def handle(
        self,
        transaction: aorta.types.ITransaction,
        handler: aorta.types.IMessageHandler,
        envelope: aorta.types.Envelope[Any]
    ) -> tuple[bool, Any]:
            async with transaction:
                dependant = get_dependant(call=handler.run, path='/')
                dependant.dependencies.extend(self.get_injectors(handler))
                try:
                    values, errors, *_, cache = await solve_dependencies(
                        request=self.request,
                        dependant=dependant,
                        body=envelope.model_dump(),
                        dependency_overrides_provider=None,
                        dependency_cache=self.dependency_cache
                    )
                    if errors:
                        raise RequestValidationError(errors)
                    self.dependency_cache.update(cache)
                    assert dependant.call is not None
                    assert callable(dependant.call)
                    # In some versions of FastAPI the envelope type is changed
                    # to the generic type (TODO).
                    values['envelope'] = envelope
                    
                    return await dependant.call(**values)
                except Drop as e:
                    self.logger.warning(
                        "Dropping %s/%s: %s (id: %s, correlation-id: %s)",
                        envelope.api_version,
                        envelope.kind,
                        e.reason or 'reason unknown',
                        envelope.metadata.uid,
                        envelope.metadata.correlation_id
                    )
                    return True, NotImplemented

    def get_injectors(self, obj: Any) -> Generator[Dependant, None, None]:
        for attname, member in inspect.getmembers(obj):
            if not isinstance(member, fastapi.params.Depends):
                continue
            def setter(attname: str, dep: Any = member):
                async def f(dep: Any = dep) -> None:
                    setattr(obj, attname, dep)
                return f
            yield get_parameterless_sub_dependant(
                 depends=fastapi.Depends(setter(attname, member)),
                 path='/'
            )


        seen: set[str] = set()
        for cls in inspect.getmro(type(obj)):
            annotations = inspect.get_annotations(cls)
            for attname, annotation in annotations.items():
                if attname in seen:
                    continue
                if get_origin(annotation) != Annotated:
                    continue
                args = get_args(annotation)
                if len(args) != 2 or not isinstance(args[1], fastapi.params.Depends):
                    continue

                def setter(attname: str, dep: Any):
                    async def f(dep: Any = dep) -> None:
                        setattr(obj, attname, dep)
                    return f
                yield get_parameterless_sub_dependant(
                    depends=fastapi.Depends(setter(attname, args[1])),
                    path='/'
                )
                seen.add(attname)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from unittest.mock import AsyncMock

import pytest

import aorta


__all__: list[str] = [
    'test_handle_success',
    'test_issue_command',
    'test_run',
    'test_run_exception_on_bound_envelope',
    'test_run_exception_on_unbound_envelope',
    'test_publish_event',
    'test_get_handlers',
]


@pytest.mark.asyncio
async def test_handle_success(
    transaction: aorta.types.ITransaction,
    message: aorta.types.Publishable,
    MessageHandler: type[aorta.types.IMessageHandler]
):
    envelope = message.envelope()
    handler = MessageHandler(
        metadata=envelope.metadata,
        publisher=transaction
    )
    assert await handler.handle(envelope.message) == 'foo' # type: ignore


def test_get_handlers(message: aorta.types.Publishable):
    envelope = message.envelope()
    handlers = aorta.get(envelope)
    assert len(handlers) == 2 # includes the sewer


@pytest.mark.asyncio
async def test_run(
    message: aorta.types.Publishable,
):
    assert aorta.run(type(message), **message.model_dump())


@pytest.mark.skip("Messages are not republished on error.")
@pytest.mark.asyncio
async def test_run_exception_on_unbound_envelope(
    message: aorta.types.Publishable,
    runner: aorta.BaseRunner,
    transport: aorta.NullTransport
):
    envelope = message.envelope()
    runner.handle = AsyncMock(side_effect=Exception)
    await runner.run(envelope)
    assert len(transport.sent) == 2, transport.sent
    assert transport.sent[0].metadata.uid == envelope.metadata.uid
    assert transport.sent[0].is_bound()


@pytest.mark.skip("Messages are not republished on error.")
@pytest.mark.asyncio
async def test_run_exception_on_bound_envelope(
    message: aorta.types.Publishable,
    runner: aorta.BaseRunner,
    transport: aorta.NullTransport
):
    envelope = message.envelope()
    runner.handle = AsyncMock(side_effect=Exception)
    await runner.run(envelope)
    assert not await runner.run(transport.sent[0])


@pytest.mark.asyncio
async def test_issue_command(
    publisher: aorta.types.IPublisher,
    transport: aorta.NullTransport,
    message: aorta.types.Publishable,
    command: aorta.Command,
    MessageHandler: type[aorta.types.IMessageHandler]
):
    envelope = message.envelope()
    async with aorta.Transaction(publisher) as tx:
        assert envelope.metadata.correlation_id is not None
        handler = MessageHandler(
            metadata=envelope.metadata,
            publisher=tx
        )
        handler.issue(command)

    assert len(transport.sent) == 1
    assert transport.sent[0].metadata.correlation_id != tx.correlation_id
    assert transport.sent[0].metadata.correlation_id == envelope.metadata.correlation_id


@pytest.mark.asyncio
async def test_publish_event(
    publisher: aorta.types.IPublisher,
    transport: aorta.NullTransport,
    message: aorta.types.Publishable,
    event: aorta.Event,
    MessageHandler: type[aorta.types.IMessageHandler]
):
    envelope = message.envelope()
    async with aorta.Transaction(publisher) as tx:
        handler = MessageHandler(
            metadata=envelope.metadata,
            publisher=tx
        )
        handler.publish(event)

    assert len(transport.sent) == 1
    assert transport.sent[0].metadata.correlation_id != tx.correlation_id
    assert transport.sent[0].metadata.correlation_id == envelope.metadata.correlation_id


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

from .droppable import Droppable
from .messageheader import MessageHeader


T = TypeVar('T')


class Envelope(MessageHeader, Droppable, Generic[T]):
    model_config = {'populate_by_name': True}

    @property
    def header(self) -> MessageHeader:
        return MessageHeader.model_validate(self.model_dump())

    @property
    def message(self) -> T:
        return self.get_message()

    def clone(self, handler_class: type[Any]):
        envelope = self.model_validate(self.model_dump())
        envelope.metadata.handler = f'{handler_class.__module__}.{handler_class.__name__}'
        return envelope

    def model_dump(self, *args: Any, **kwargs: Any) -> Any:
        kwargs.setdefault('by_alias', True)
        return super().model_dump(*args, **kwargs)
    
    def get_message(self) -> T:
        raise NotImplementedError
    
    def is_bound(self) -> bool:
        return self.metadata.handler is not None

    def is_known(self) -> bool:
        """Return a boolean if the enclosed message is known."""
        return True

    def is_namespaced(self) -> bool:
        """Return a boolean indicating if the envelope is namespaced."""
        return bool(self.metadata.namespace)

    def wants(self, handler_class: type[Any]) -> bool:
        return self.metadata.handler == f'{handler_class.__module__}.{handler_class.__name__}'


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import logging.config
import time
from queue import Queue
from typing import cast
from typing import Any
from typing import TypeVar

from libcanonical.runtime import MainProcess # type: ignore

from aorta import Command
from aorta import Event
from aorta import MessagePublisher
from aorta import NullTransport
from aorta.types import Acknowledgable
from aorta.types import Envelope
from aorta.types import IMessageHandler
from aorta.types import IPublisher
from aorta.types import IRunner
from aorta.provider import Provider
from aorta.pool import ThreadWorker

E = TypeVar('E', bound=Command|Event)


class BaseConsumer(MainProcess):
    __module__: str = 'aorta.ext.consumer'
    concurrency: int = 1
    debug: bool = False
    interval = 0.25
    logger: logging.Logger = logging.getLogger('aorta.transport.ingress')
    max_messages: int = 0
    metrics: logging.Logger = logging.getLogger('aorta.metrics')
    num_workers: int = 1
    queue: Queue[tuple[Acknowledgable, Envelope[Any], set[type[IMessageHandler]]]]
    workers: list[ThreadWorker]

    @property
    def idle(self):
        return time.monotonic() - self.last_message

    def __init__(
        self,
        name: str,
        workers: int = 1,
        concurrency: int = 1,
        max_idle: int = 0,
        max_runtime: float = 0.0,
        max_messages: int = 0,
        loglevel: str = 'INFO',
        subscription: str | None = None,
        transport_loglevel: str = 'WARNING',
        **kwargs: Any
    ):
        super().__init__(name=name)
        self.concurrency = concurrency
        self.loglevel = loglevel
        self.max_idle = max_idle
        self.max_runtime = max_runtime
        self.max_messages = max_messages
        self.max_workers = workers
        self.queue = Queue()
        self.received = 0
        self.subscription = subscription
        self.transport_loglevel = transport_loglevel
        self.workers = []

    def accept_message(self, frame: Acknowledgable):
        if self.must_exit:
            return
        envelope = None
        try:
            envelope = Provider.loads(frame.data)
            if envelope is None:
                raise ValueError("Unable to parse message from incoming data.")
        except Exception as e:
            self.logger.warning("Caught fatal %s", repr(e))
            return
        assert envelope is not None
        if isinstance(envelope, Envelope):
            self.on_message_received(frame, cast(Envelope[Any], envelope))
        else:
            self.logger.warning(
                "Received unknown Aorta message (kind: %s, uid: %s, id: %s)",
                envelope.kind,
                envelope.metadata.uid[:6],
                envelope.metadata.correlation_id[:6]
            )

    def can_accept(self):
        return not any([
            (self.received >= self.max_messages) and (self.max_messages > 0),
            self.max_runtime and (self.age > self.max_runtime) and self.received == 0,
            (self.idle > self.max_idle) and self.max_idle > 0
        ])

    def configure(self, reloading: bool = False):
        self.logger.debug("Running with %s workers", self.num_workers)
        if not reloading:
            self.last_message = time.monotonic()
            self.publisher = self.get_publisher()
            for _ in range(self.max_workers):
                self.workers.append(self.initialize_worker())
        if self.max_messages:
            self.logger.info("Worker will consume at most %s messages", self.max_messages)

    def configure_worker(self) -> None:
        logging.config.dictConfig(dict(self.get_logging_config()))

    def get_publisher(self) -> IPublisher:
        return MessagePublisher(
            transport=NullTransport()
        )

    def get(self):
        return self.queue.get(block=False)

    def get_runner(self, publisher: IPublisher) -> IRunner:
        raise NotImplementedError

    def initialize_worker(self) -> ThreadWorker:
        raise NotImplementedError

    def route(self, envelope: Envelope[E]) -> list[str]:
        """Return a list of strings indicating the topics that the
        `envelope` must be sent to. Subclasses must override this
        method.
        """
        raise NotImplementedError

    def main_event(self) -> None:
        must_exit = not self.can_accept()
        if not must_exit:
            self.pull()
            return
        self.logger.warning(
            "Shutdown condition satisfied (received: %s, age: %.02fs, idle: %.02fs).",
            self.received,
            self.age,
            self.idle
        )
        self.stop()
        for worker in self.workers:
            worker.stop()
            self.logger.debug("Closing worker (id: %s)", worker.ident)
        t0 = time.monotonic()
        while any([w.is_alive() for w in self.workers]):
            time.sleep(1)
            dt = (time.monotonic() - t0)
            if dt > 180:
                self.logger.warning("Workers did not stop after %.02f seconds.", dt)

    def on_message_received(
        self,
        frame: Acknowledgable,
        envelope: Envelope[Any],
    ):
        self.queue.put((frame, envelope, Provider.get(envelope)))
        self.last_message = time.monotonic()

    def pull(self):
        pass


# Copyright (C) 2016-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
import time
from typing import Any

from google.api_core.exceptions import InvalidArgument
from google.api_core.exceptions import NotFound
from google.cloud.pubsub_v1 import SubscriberClient # type: ignore
from libcanonical.utils.logging import LoggingConfigDict # type: ignore

import aorta
from aorta import MessagePublisher
from aorta.ext.consumer import BaseConsumer
from aorta.runners import SequentialRunner
from aorta.types import Acknowledgable
from aorta.types import IRunner
from aorta.types import Envelope
from aorta.types import IPublisher
from aorta.types import ITransport
from .googletransport import GoogleTransport
from ._frame import Frame
from ._pubsubworker import PubsubWorker


class TopicListener(BaseConsumer):
    debug = True
    project: str | None = None
    publisher: IPublisher
    transport_class: type[GoogleTransport] = GoogleTransport

    @property
    def prefetch(self):
        return self.num_workers * self.concurrency

    def configure(self, reloading: bool = False):
        self.runner = SequentialRunner()
        if self.project is None:
            self.project = os.environ['GOOGLE_PUBSUB_PROJECT']
        self.subscriber = SubscriberClient()
        if not self.subscription:
            self.subscription = os.environ['GOOGLE_PUBSUB_SUBSCRIPTION']
        self.subscription = self.subscriber.subscription_path(self.project, self.subscription)
        super().configure(reloading=reloading)
        self.logger.debug("Established connection to subscription %s", self.subscription)

    def get_logging_config(self) -> LoggingConfigDict:
        config = super().get_logging_config()
        config['loggers']['aortra.metrics'] = {
            'handlers': ['google-cloud'],
            'level': 'INFO',
            'propagate': False
        }
        return config

    def get_publisher(self) -> IPublisher:
        return MessagePublisher(transport=self.get_transport())

    def get_runner(self, publisher: IPublisher) -> IRunner:
        return self.runner

    def route(self, envelope: Envelope[Any]) -> list[str]:
        if not isinstance(envelope.message, aorta.Command):
            return []
        return ['dataverse.commands']

    def get_transport(self) -> ITransport:
        if self.project is None:
            raise self.ConfigurationFailure(
                "Project was not provided and could not be inferred from "
                "the environment."
            )
        return self.transport_class(
            project=self.project,
            prefix=self.name
        )

    def initialize_worker(self) -> PubsubWorker:
        return PubsubWorker(pool=self, limit=self.concurrency)

    def pull(self) -> None:
        if self.must_exit:
            return
        n = max(self.prefetch - self.queue.qsize(), 0)
        if n == 0:
            self.logger.debug("Suspending message pull (running: %s)", self.queue.qsize())
            return
        try:
            response = self.subscriber.pull(  # type: ignore
                subscription=self.subscription,
                max_messages=self.prefetch,
                return_immediately=True
            )
            if not response.received_messages:
                return
            frames: list[Acknowledgable] = []
            ack_ids: list[str] = []
            for message in response.received_messages:
                frame = Frame(message)
                if not self.can_accept():
                    break
                ack_ids.append(message.ack_id)
                frames.append(frame)
                self.accept_message(frame)
            self.subscriber.acknowledge( # type: ignore
                subscription=self.subscription,
                ack_ids=ack_ids
            )
        except (InvalidArgument, NotFound) as e:
            self.logger.critical(e.message) # type: ignore
            time.sleep(5)


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import ParamSpec
from typing import Protocol
from typing import TypeVar

from .command import Command
from .event import Event
from .publishable import Publishable


C = TypeVar('C', bound=Command)
E = TypeVar('E', bound=Event)
D = TypeVar('D')
P = ParamSpec('P')
T = TypeVar('T', bound='ITransaction')


class ITransaction(Protocol):
    __module__: str = 'aorta.types'
    correlation_id: str

    def rollback(self) -> None:
        ...

    def pending(self) -> list[Publishable]:
        ...

    def publish(
        self,
        message: Publishable,
        correlation_id: str | None = None,
        audience: set[str] | None = None
    ):
        ...

    def issue(
        self,
        c: Callable[P, C] | C,
        correlation_id: str | None = None,
        audience: set[str] | None = None,
        *args: P.args,
        **kwargs: P.kwargs
    ):
        ...

    def notify(
        self,
        e: Callable[P, E] | E,
        correlation_id: str | None = None,
        audience: set[str] | None = None,
        *args: P.args,
        **kwargs: P.kwargs
    ):
        ...

    async def commit(self) -> None: ...
    async def send(self, message: Publishable) -> None: ...
    async def __aenter__(self: T) -> T: ...
    async def __aexit__(self, cls: type[BaseException] | None, *args: Any): ...


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from .acknowledgable import Acknowledgable
from .envelope import Envelope
from .imessagehandler import IMessageHandler
from .ipublisher import IPublisher
from .irunner import IRunner


class IPool(Protocol):
    __module__: str = 'aorta.types'

    def can_accept(self) -> bool:
        ...

    def get(self) -> tuple[Acknowledgable, Envelope[Any], set[type[IMessageHandler]]]:
        ...

    def get_publisher(self) -> IPublisher:
        ...

    def get_runner(self, publisher: IPublisher) -> IRunner:
        ...


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import ClassVar
from typing import Self
from typing import TypeVar

from .eventenvelope import EventEnvelope
from .message import Message

T = TypeVar('T', bound='Event')


class Event(Message):
    __envelope__ = EventEnvelope
    __module__: str = 'aorta.types'
    __message_attr__ = 'data'
    __message_type__ = 'cochise.io/event'
    __registry__: ClassVar[dict[tuple[str, str], type[Self]]] = {}


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias
from typing import Union

from .acknowledgable import Acknowledgable
from .command import Command
from .commandenvelope import CommandEnvelope
from .drop import Drop
from .envelope import Envelope
from .event import Event
from .eventenvelope import EventEnvelope
from .handlerexception import HandlerException
from .imessagehandler import IMessageHandler
from .ipool import IPool
from .ipublisher import IPublisher
from .itransaction import ITransaction
from .itransport import ITransport
from .irunner import IRunner
from .message import Message
from .messagehandlertype import MessageHandlerType
from .messageheader import MessageHeader
from .messagemetadata import MessageMetadata
from .publishable import Publishable


__all__: list[str] = [
    'Acknowledgable',
    'Command',
    'CommandEnvelope',
    'Drop',
    'Envelope',
    'Event',
    'EventEnvelope',
    'HandlerException',
    'IMessageHandler',
    'IPool',
    'IPublisher',
    'IRunner',
    'ITransaction',
    'ITransport',
    'Message',
    'MessageHandlerType',
    'MessageHeader',
    'MessageMetadata',
    'Publishable',
]

MessageType: TypeAlias = Union[
    Command,
    Event
]


# Copyright (C) 2020-2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`GoogleTransport`."""
from typing import Any

from aorta.types import Envelope
from .pubsubtransport import PubsubTransport


class GoogleTransport(PubsubTransport):
    cc: list[str] | None
    prefix: str
    service_name: str | None

    def __init__(
        self,
        *,
        project: str,
        prefix: str,
        service_name: str | None = None,
        cc: list[str] | None = None
    ):
        self.cc = cc
        self.prefix = prefix
        self.service_name = service_name
        super().__init__(
            project=project,
            topic=self.topic_factory,
            retry_topic=f'{self.prefix}.retry.{service_name}'
        )

    def topic_factory(self, envelope: Envelope[Any]) -> list[str]:
        if envelope.is_event():
            topics: list[str] = [
                f'{self.prefix}.events.{envelope.kind}'
            ]
            if self.cc is not None:
                topics.extend(self.cc)
        elif envelope.is_command():
            topics = [f'{self.prefix}.commands']
            if self.service_name is not None:
                topics = [f'{self.prefix}.commands.{self.service_name}']
        else:
            raise NotImplementedError(envelope.type)
        return topics


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from google.pubsub_v1 import ReceivedMessage

from aorta.types import Acknowledgable


class Frame(Acknowledgable):

    def __init__(self, frame: ReceivedMessage) -> None:
        self.data = frame.message.data
        self.frame = frame
        self.message_id = frame.message.message_id

    def ack(self):
        pass

    def nack(self):
        pass


# Copyright (C) 2016-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any
from typing import Callable

from aorta.types import Envelope
from aorta.types import IPublisher
from aorta.types import IRunner
from aorta.types import Message
from aorta.types import MessageHeader
from .messagehandler import MessageHandler
from .transaction import Transaction


class LocalPublisher(IPublisher):
    """A publisher implementation that publishes and runs messages in
    the local process and thread.
    """
    __module__: str = 'aorta'

    def __init__(
        self,
        provider: Callable[[Envelope[Any] | MessageHeader, bool], set[type[MessageHandler[Any]]]],
        runner: IRunner
    ) -> None:
        self.provider = provider
        self.runner = runner

    def begin(self):
        return Transaction(publisher=self)

    async def publish(
        self,
        message: Message,
        **kwargs: Any
    ) -> None:
        envelope = message.envelope()
        await self.runner.run(
            publisher=self,
            envelope=envelope,
            handlers=self.provider(envelope, False),
        )

    async def send(
        self,
        messages: list[Envelope[Any]],
        is_retry: bool = False
    ):
        await asyncio.gather(*[
            self.runner.run(
                publisher=self,
                envelope=m,
                handlers=self.provider(m, False),
            )
            for m in messages
        ])